/**
 * @file include/gb2D.hpp
 *
 * ghulbusGraphics - The Ghulbus Graphics Library
 *
 * @brief gbGraphics 2D drawing routines
 *
 * @version 0.1
 * @author Andreas Weis
 *
 */

#ifndef _GHULBUSGRAPHICS_2D_HPP_INCLUDE_GUARD_
#define _GHULBUSGRAPHICS_2D_HPP_INCLUDE_GUARD_

#include "gbException.hpp"
#include "gbColor.hpp"
#include "gbTexture.hpp"

namespace GhulbusGraphics {

	/** 2d drawing
	 */
	class gb2D {
		GBCOLOR*           m_Data;			///< pointer to surface
		int                m_width;			///< width of the surface
		int                m_height;		///< height of the surface
		int                m_pitch;			///< number of pixels in a row
		bool               m_isLocked;		///< true between BeginScene() and EndScene()
		GBCOLOR            m_clearcolor;	///< color used for clearing
		gbTexture*         m_Texture;		///< stores the texture object if a texture is locked
		Ghulbus::gbException::GBERROR m_errorstate;		///< internal errorstate
	public:
		/** Constructor
		 */
		gb2D();
		/** Constructor
		 * @param[in] data Pointer to the surface that is to be drawn to
		 * @param[in] width Width of the surface
		 * @param[in] height Height of the surface
		 * @param[in] pitch Number of pixels per row
		 * @throw Ghulbus::gbException GB_ILLEGALPARAMETER
		 */
		gb2D(GBCOLOR* data, int width, int height, int pitch);
		/** Constructor
		 * @param[in] tex A gbTexture that is cleared for locking
		 * @throw Ghulbus::gbException GB_ILLEGALPARAMETER
		 */
		gb2D(gbTexture* tex);
		/** Destructor
		 */
		~gb2D();
		/** Acquires a surface for drawing
		 * @param[in] data Pointer to the surface that is to be drawn to
		 * @param[in] width Width of the surface
		 * @param[in] height Height of the surface
		 * @param[in] pitch Number of pixels per row
		 * @throw Ghulbus::gbException GB_ILLEGALPARAMETER, GB_INVALIDCONTEXT
		 */
		void BeginScene(GBCOLOR* data, int width, int height, int pitch);
		/** Acquire a gbTexture surface for drawing
		 * @param[in] tex A gbTexture that is cleared for locking
		 * @throw Ghulbus::gbException GB_ILLEGALPARAMETER, GB_INVALIDCONTEXT
		 */
		void BeginScene(gbTexture* tex);
		/** Releases a surface from drawing
		 * @throw Ghulbus::gbException GB_INVALIDCONTEXT
		 */
		void EndScene();
		/** Is gb2D ready for drawing
		 * @return true if called between valid BeginScene() and EndScene()
		 */
		bool IsLocked();
		/** Get the width of the locked surface
		 * @return The surface width in pixels, -1 if no surface is present
		 */
		int GetWidth();
		/** Get the height of the locked surface
		 * @return The surface height in pixels, -1 if no surface is present
		 */
		int GetHeight();
		/** Get the internal errorstate
		 * @return The internal errorstate; If != GB_OK, an error did occur
		 * @note Each call of GetErrorState() automatically resets the
		 *       errorstate back to GB_OK. The errorstate has no effect on
		 *       the behavior of the gb2D!
		 */
		Ghulbus::gbException::GBERROR GetErrorState();
		/** Set the clear color
		 * @param[in] c The new clear color
		 */
		void SetClearColor(GBCOLOR c);

		/** @name Drawing functions:
		 * @{
		 */
		/** Clear the entire surface with the current clear color
		 * @throw Ghulbus::gbException GB_INVALIDCONTEXT
		 */
		void Clear();
		/** Draw a single pixel
		 * @param[in] x X coordinate of the pixel on the surface
		 * @param[in] y Y coordinate of the pixel on the surface
		 * @param[in] c New color of the pixel
		 * @throw Ghulbus::gbException GB_INVALIDCONTEXT
		 * @note Sets errorstate to GB_ILLEGALPARAMETER if coordinates
		 *       are out of bounds.
		 */
		void PutPixel(int x, int y, GBCOLOR c);
		/** Draw a line between two points
		 * @param[in] startX X coordinate of the start point
		 * @param[in] startY Y coordinate of the start point
		 * @param[in] endX X coordinate of the end point
		 * @param[in] endY Y coordinate of the end point
		 * @param[in] c Color of the line
		 * @throw Ghulbus::gbException GB_INVALIDCONTEXT
		 * @note Sets errorstate to GB_ILLEGALPARAMETER if coordinates
		 *       are out of bounds.
		 */
		void Line(int startX, int startY, int endX, int endY, GBCOLOR color);
		/** Draw a horizontal line between two points
		 * @param[in] startX X coordinate of the start point
		 * @param[in] endX X coordinate of the end point
		 * @param[in] y Y coordinate of the line
		 * @param[in] c Color of the line
		 * @throw Ghulbus::gbException GB_INVALIDCONTEXT
		 * @note Sets errorstate to GB_ILLEGALPARAMETER if any
		 *       pixels were out of bounds.
		 */
		void HLine(int startX, int endX, int y, GBCOLOR color);
		/** Draw a horizontal line with interpolated colors between two points
		 * @param[in] startX X coordinate of the start point
		 * @param[in] endX X coordinate of the end point
		 * @param[in] y Y coordinate of the line
		 * @param[in] c1 Color of the start point
		 * @param[in] c2 Color of the end point
		 * @throw Ghulbus::gbException GB_INVALIDCONTEXT
		 * @note Sets errorstate to GB_ILLEGALPARAMETER if any
		 *       pixels were out of bounds.
		 */
		void HLineGradient(int startX, int endX, int y, GBCOLOR c1, GBCOLOR c2);
		/** Draw a circle
		 * @param[in] posX X coordinate of the circle center
		 * @param[in] posY Y coordinate of the circle center
		 * @param[in] radius Circle radius
		 * @param[in] c Color of the circle
		 * @throw Ghulbus::gbException GB_INVALIDCONTEXT
		 * @note Sets errorstate to GB_ILLEGALPARAMETER if any
		 *       pixels were out of bounds.
		 */
		void Circle(int posX, int posY, int radius, GBCOLOR color);
		/** Draw a rectangle
		 * @param[in] left Left limit of the rectangle
		 * @param[in] top Upper limit of the rectangle
		 * @param[in] right Right limit of the rectangle
		 * @param[in] bottom Lower limit of the rectangle
		 * @param[in] c Color of the rectangle
		 * @throw Ghulbus::gbException GB_INVALIDCONTEXT
		 * @note Sets errorstate to GB_ILLEGALPARAMETER if coordinates
		 *       are out of bounds.
		 */
		void Rectangle(int left, int top, int right, int bottom, GBCOLOR color);
		/** Draw a rectangle with interpolated colors
		 * @param[in] left Left limit of the rectangle
		 * @param[in] top Upper limit of the rectangle
		 * @param[in] right Right limit of the rectangle
		 * @param[in] bottom Lower limit of the rectangle
		 * @param[in] ul Color of the upper left corner point
		 * @param[in] ur Color of the upper right corner point
		 * @param[in] ll Color of the lower left corner point
		 * @param[in] lr Color of the lower right corner point
		 * @throw Ghulbus::gbException GB_INVALIDCONTEXT
		 * @note Sets errorstate to GB_ILLEGALPARAMETER if any
		 *       pixels were out of bounds.
		 */
		void RectangleGradient(int left, int top, int right, int bottom, 
							   GBCOLOR ul, GBCOLOR ur, GBCOLOR ll, GBCOLOR lr);
		/** Draw a filled rectangle
		 * @param[in] left Left limit of the box
		 * @param[in] top Upper limit of the box
		 * @param[in] right Right limit of the box
		 * @param[in] bottom Lower limit of the box
		 * @param[in] c Color of the box
		 * @throw Ghulbus::gbException GB_INVALIDCONTEXT
		 */
		void Box(int left, int top, int right, int bottom, GBCOLOR color);
		/** Draw a filled rectangle with interpolated colors
		 * @param[in] left Left limit of the box
		 * @param[in] top Upper limit of the box
		 * @param[in] right Right limit of the box
		 * @param[in] bottom Lower limit of the box
		 * @param[in] ul Color of the upper left corner point
		 * @param[in] ur Color of the upper right corner point
		 * @param[in] ll Color of the lower left corner point
		 * @param[in] lr Color of the lower right corner point
		 * @throw Ghulbus::gbException GB_INVALIDCONTEXT
		 * @note Sets errorstate to GB_ILLEGALPARAMETER if any
		 *       pixels were out of bounds.
		 */
		void BoxGradient(int left, int top, int right, int bottom, 
						 GBCOLOR ul, GBCOLOR ur, GBCOLOR ll, GBCOLOR lr);
		/// @}
	private:
		/** Helper function: Throws an exception if m_isLocked is false
		 * @throws Ghulbus::gbException GB_INVALIDCONTEXT
		 */
		void TestLock();
	};
	/*-------------------------------------
	 | Extensive Documentation:           |
	 *------------------------------------*/
	/** 
	 * @class gb2D
	 *
	 * @details The gb2D class can be used for all kinds of 2d drawing and
	 * plotting. A gb2D object provides only the algorithms for drawing, but
	 * stores no actual data. Therefore, a surface must be locked with gb2D
	 * before any drawing can be performed. A surface can be locked by either
	 * calling BeginScene() or by passing it to the constructor.
	 * Remember that gb2D is actually locking the surfaces, so they must be
	 * released before they can be rendered. Therefore all calls to gb2D must
	 * occur outside of the gbGraphicsManager::BeginScene()/EndScene() block.
	 * Furthermore, gb2D keeps no history of the images it drew. Once a texture
	 * was cleared, all operations executed by gb2D have to be performed again.
	 * Keep this is mind when drawing to textures that are automatically cleared
	 * each frame, like the Overlay of gbGraphicsManager.
	 *
	 */
}

#endif
