/**
 * @file include/gbGraphics.hpp
 *
 * ghulbusGraphics - The Ghulbus Graphics Library
 *
 * @brief gbGraphics Main Header File
 *
 * @version 0.1
 * @author Andreas Weis
 *
 */

#ifndef _GHULBUSGRAPHICS_HPP_INCLUDE_GUARD_
#define _GHULBUSGRAPHICS_HPP_INCLUDE_GUARD_

#include "gbException.hpp"
#include "gbColor.hpp"
#include "gbTexture.hpp"
#include "gbGraphicsObject.hpp"
#include "gbSprite.hpp"
#include <vector>
#include <map>

#ifdef _GHULBUSGRAPHICS__USE_SDL_
#	ifdef WIN32
#		include <SDL.h>
#		define WIN32_LEAN_AND_MEAN
#		include <windows.h>
#	else
#		include <SDL/SDL.h>
#	endif
#	include <GL/glew.h>
#	include <GL/gl.h>
#	include <GL/glu.h>
#endif

#ifdef _GHULBUSGRAPHICS__USE_D3D9_
#	ifndef WIN32
#		undef _GHULBUSGRAPHICS__USE_D3D9_
#	else
#		define WIN32_LEAN_AND_MEAN
#		include <windows.h>
#		include <d3d9.h>
#		include <d3dx9.h>
#	endif
#endif

/** Root namespace of the gbGraphics library
 */
namespace GhulbusGraphics {

	/** Root object of gbGraphics
	 */
	class gbGraphicsManager {
	protected:
		int m_scrWidth;									///< viewport width in pixels
		int m_scrHeight;								///< viewport height in pixels
		bool m_isFullscreen;							///< true if running fullscreen
		bool m_isRendering;								///< true between BeginScene() and EndScene()
		bool m_done;									///< true after SetDone() was called
		GBCOLOR m_clearcolor;							///< current clearcolor
		std::vector<gbGraphicsObject*> m_ObjectList;	///< list of all registered gbGraphicsObject
		gbTexture*                     m_Overlay;		///< a texture overlay rendered on each frame
	protected:
		/** Constructor
		 * Fills base members with default values
		 */
		gbGraphicsManager();
	public:
		/** Destructor
		 */
		virtual ~gbGraphicsManager();
		/** Responsible for opening the scene for manipulation
		 * @note BeginScene also presents the framebuffer of the last frame
		 *       and clears the scene for the next frame.
		 * @throw Ghulbus::gbException GB_INVALIDCONTEXT
		 */
		virtual void BeginScene()=0;
		/** Responsible for locking the scene before presentation
		 * @throw Ghulbus::gbException GB_INVALIDCONTEXT
		 */
		virtual void EndScene()=0;
		/** Determines whether the scene is open for manipulation
		 * @return true if called between BeginScene() and EndScene()
		 */
		bool IsRendering();
		/** Sets the framebuffer clear color
		 * @param[in] c The new clear color
		 */
		virtual void SetClearColor(GBCOLOR c);
		/** Clear the framebuffer
		 * @throw Ghulbus::gbException GB_INVALIDCONTEXT
		 * @note Depth and stencil buffer are left untouched
		 */
		virtual void Clear()=0;
		/** Get the current viewport width
		 * @return The viewport's width in pixels
		 */
		int GetScreenWidth();
		/** Get the current viewport height
		 * @return The viewport's height in pixels
		 */
		int GetScreenHeight();
		/** Determine whether the current viewport is fullscreen
		 * @return true if fullscreen, false if windowed
		 */
		bool IsFullscreen();
		/** Get the status of the m_done flag
		 * @return true if SetDone() was called before, false otherwise
		 */
		bool IsDone();
		/** Set the m_done flag
		 */
		void SetDone();
		/** Get the texture object wrapping the Overlay texture
		 * @return The Overlay texture
		 * @throw Ghulbus::gbException GB_INVALIDCONTEXT - if the manager is rendering 
		 *                             or the Overlay is locked;
		 */
		virtual gbTexture* GetOverlaySurface();
		/** ???
		 */
		virtual void Render(unsigned int dT)=0;
		/** A message callback that is expected to be called once every frame
		 */
		virtual void ProcessMessages()=0;
		/** Create a Sprite of the specified size
		 * @param[in] width the requested image width in pixels
		 * @param[in] height the requested image height in pixels
		 * @return A gbSprite object
		 */
		virtual gbSprite* CreateSprite(int width, int height)=0;
	private:
		gbGraphicsManager(gbGraphicsManager const&);				///< Private copy constructor (not implemented)
		gbGraphicsManager& operator=(gbGraphicsManager const&);		///< Private copy assignment (not implemented)
	};

#ifdef _GHULBUSGRAPHICS__USE_SDL_
	class gbGraphicsManager_SDL: public gbGraphicsManager {
	private:
		SDL_Surface* m_Screen;
		GLenum       m_glError;
		bool		 m_npot_textures_supported;
	public:
		gbGraphicsManager_SDL(int scrWidth, int scrHeight, bool fullscreen);
		virtual ~gbGraphicsManager_SDL();
		void BeginScene();
		void EndScene();
		void Clear();
		virtual void SetClearColor(GBCOLOR c);
		void Render(unsigned int dT);
		gbSprite* CreateSprite(int width, int height);
		void ProcessMessages();
	private:
		/** Helper function: Set the initial OpenGL state
		 */
		void InitOpenGL();
		/** Helper function: Check required extensions
		 */
		void LoadOpenGLExtensions();
		/** Helper function: Build the Overlay texture
		 */
		void BuildOverlay();
		/** Helper function: Render the Overlay texture
		 */
		void RenderOverlay();
	};
#endif

#ifdef _GHULBUSGRAPHICS__USE_D3D9_
	class gbGraphicsManager_D3D9: public gbGraphicsManager {
	private:
		static std::map<HWND, gbGraphicsManager_D3D9*> m_CallbackMap;
		HWND	              m_hWnd;
		HRESULT	              m_hres;
		HINSTANCE             m_hInst;
		LPDIRECT3D9           m_lpD3D;
		LPDIRECT3DDEVICE9     m_lpDev;
		wchar_t*              m_lpClassName;
		unsigned int          m_AdapterRecommendation;
		D3DPRESENT_PARAMETERS m_d3dpp;
		LPDIRECT3DTEXTURE9    m_lpSurface;
		LPD3DXSPRITE          m_lpSurfaceSprite;
	public:
		/** Constructor
		 * @note To ensure thread safety, constructor calls must not be parallel!
		 */
		gbGraphicsManager_D3D9(int scrWidth, int scrHeight, bool fullscreen);
		virtual ~gbGraphicsManager_D3D9();
		void BeginScene();
		void EndScene();
		void Clear();
		void Render(unsigned int dT);
		void ProcessMessages();
		gbSprite* CreateSprite(int width, int height);
	private:
		/** Actual member callback; Responsible for message processing
		 */
		virtual bool ProcessWindowMessage(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam, LRESULT* lResult);
		/** Static callback for message processing
		 * Responsible for delivering messages to ProcessWindowMessage()
		 */
		static LRESULT CALLBACK MsgCallback(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
		/** Helper function: Bootstrap callback
		 * Responsible for setting up environment for message processing by ProcessWindowMessage()
		 */
		static LRESULT CALLBACK BootstrapMsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
		/** Helper function: Create a Win32 window
		 */
		HWND CreateWin32Window(int winWidth, int winHeight);
		/** Helper function: Enumerate available D3D adapters and modes
		 */
		bool Enumerate(D3DFORMAT format, unsigned int resX, unsigned int resY);
		/** Helper function: Create a D3D device
		 */
		void CreateD3DDevice();
		/** Helper function: Create drawing surface texture and sprite
		 */
		void CreateD3DDrawingSurface();
	};
#endif
}

#endif
