/**
 * @file include/gbTexture.hpp
 *
 * ghulbusGraphics - The Ghulbus Graphics Library
 *
 * @brief gbGraphics Texture header file
 *
 * @version 0.1
 * @author Andreas Weis
 *
 */

#ifndef _GHULBUSGRAPHICS_TEXTURE_HPP_INCLUDE_GUARD_
#define _GHULBUSGRAPHICS_TEXTURE_HPP_INCLUDE_GUARD_

#include "gbException.hpp"
#include "gbColor.hpp"

#ifdef _GHULBUSGRAPHICS__USE_SDL_
#	ifdef WIN32
#		include <SDL.h>
#		define WIN32_LEAN_AND_MEAN
#		include <windows.h>
#	else
#		include <SDL/SDL.h>
#	endif
#	include <GL/glew.h>
#	include <GL/gl.h>
#	include <GL/glu.h>
#endif

#ifdef _GHULBUSGRAPHICS__USE_D3D9_
#	ifndef WIN32
#		undef _GHULBUSGRAPHICS__USE_D3D9_
#	else
#		define WIN32_LEAN_AND_MEAN
#		include <windows.h>
#		include <d3d9.h>
#		include <d3dx9.h>
#	endif
#endif

namespace GhulbusGraphics {

	/** A wrapper for stored texture data
	 */
	class gbTexture {
	protected:
		GBCOLOR* m_Data;		///< a pointer to the texture's image data (only valid if locked)
		int      m_width;		///< texture width in pixels
		int      m_height;		///< texture height in pixels
		bool     m_isLocked;	///< flag indicating whether the texture is locked
	protected:
		/** Constructor
		 * @note The actual signature of the constructor depends on the implementation.
		 *       Therefore, using a factory is recommended for instantiation (e.g. gbGraphicsManager)
		 */
		gbTexture();
	public:
		/** Destructor
		 */
		virtual ~gbTexture();
		/** Get the texture width
		 * @return The texture width in pixels
		 */
		int GetWidth();
		/** Get the texture height
		 * @return The texture height in pixels
		 */
		int GetHeight();
		/** Lock the texture and acquire a pointer to the image data
		 * @param[out] pDara A pointer to the image data
		 * @param[out] pPitch The number of entries in a row
		 * @note The value of pPitch is not necessarily equal to the image width!
		 */
		virtual void LockTexture(GBCOLOR** pData, int* pPitch)=0;
		/** Unlock a texture and write back the new data
		 * @note All previously acquired handles to the image data (e.g. through
		 *       calling LockTexture()) become invalid!
		 */
		virtual void UnlockTexture()=0;
		/** Check whether the texture is currently locked
		 * @return true if the texture is currently locked, false otherwise
		 */
		bool IsLocked();
	};

#ifdef _GHULBUSGRAPHICS__USE_SDL_
	class gbTexture_SDL: public gbTexture {
	private:
		GLuint   m_textureId;
		GBCOLOR* m_Data;
		int      m_texWidth;
		int      m_texHeight;
	public:
		gbTexture_SDL(GLuint texture_id, int width, int height, bool npot_supported);
		virtual ~gbTexture_SDL();
		virtual void LockTexture(GBCOLOR** pData, int* pPitch);
		virtual void UnlockTexture();
		virtual int GetTextureWidth();
		virtual int GetTextureHeight();
		GLuint GetTextureId();
		int GetPOTSize(int npot_size);
	};
#endif

#ifdef _GHULBUSGRAPHICS__USE_D3D9_
	class gbTexture_D3D9: public gbTexture {
	private:
		LPDIRECT3DTEXTURE9 m_lpTex;
	public:
		gbTexture_D3D9(LPDIRECT3DTEXTURE9 texture);
		virtual ~gbTexture_D3D9();
		virtual void LockTexture(GBCOLOR** pData, int* pPitch);
		virtual void UnlockTexture();
		virtual LPDIRECT3DTEXTURE9 GetD3D9Texture();
	};
#endif

	/*-------------------------------------
	 | Extensive Documentation:           |
	 *------------------------------------*/
	/** 
	 * @class gbTexture
	 *
	 * @details The texture class is an abstract wrapper for texture data
	 * stored by gbGraphics. The texture class merely organizes the texture
	 * information, it does not manage them! All ressources stored by a
	 * a gbTexture must be freed either by the object which created the gbTexture
	 * or by an object which required a texture as a construction parameter.
	 * The only responsibility of this class is therefore the maintenance
	 * of texture specific data, like the associated image data and dimensions.
	 * The most important aspect of the gbTexture is its ability to lock
	 * a texture between renderings and change the associated image data.
	 * @note Textures usually have no immediate representation in a scene,
	 * except the objects that they apply to. Use a gbSprite if you want to
	 * render image data in your scene.
	 *
	 * @section gbtexture_d3d9 Implementation Details Direct3D9
	 * In D3D9 the gbTexture takes a fully initialized LPDIRECT3DTEXTURE9 as
	 * a construction parameter. Locking is performed by invoking the
	 * specific routines of the D3D texture.
	 * 
	 * @section gbtexture_sdl Implementation Details SDL
	 * The OpenGL implementation adds an additional responsibility to the
	 * class. The calls to glTexImage2D() and glTexParameter() which are
	 * vital for texture creation in OpenGL are executed in the constructor
	 * of gbTexture rather than by the instantiating object.
	 * Since OpenGL provides little more representation of its textures than
	 * a numerical id, the implementation is a little more complicated than
	 * its Direct3D equivalent.
	 * OpenGL requires a special extension (GLEW_ARB_texture_non_power_of_two)
	 * for arbitrary sized textures. This extension is checked by the manager
	 * and its availability is passed in form of a construction parameter.
	 * If this flag is not available, the texture is automatically enlarged to
	 * the next power-of-two size that can hold the requested size.
	 *
	 */
}

#endif
