#include "../include/gb2D.hpp"

namespace GhulbusGraphics {
	static const int N_CHANNELS = sizeof(GBCOLOR) / sizeof(GBCOLOR_COMPONENT);

	gb2D::gb2D() 
		:m_Data(NULL), m_width(-1), m_height(-1), m_pitch(-1), m_isLocked(false), 
		 m_clearcolor(GBCOLOR32::ARGB(0, 0, 0, 0)), m_Texture(NULL),
		 m_errorstate(Ghulbus::gbException::GB_OK)
	{
		;
	}
	gb2D::gb2D(GBCOLOR* data, int width, int height, int pitch)
		:m_Data(NULL), m_width(-1), m_height(-1), m_pitch(-1), m_isLocked(false),
		 m_clearcolor(GBCOLOR32::ARGB(0, 0, 0, 0)), m_Texture(NULL),
		 m_errorstate(Ghulbus::gbException::GB_OK)
	{
		BeginScene(data, width, height, pitch);
	}
	gb2D::gb2D(gbTexture* tex) 
		:m_Data(NULL), m_width(-1), m_height(-1), m_pitch(-1), m_isLocked(false),
		 m_clearcolor(GBCOLOR32::ARGB(0, 0, 0, 0)), m_Texture(NULL),
		 m_errorstate(Ghulbus::gbException::GB_OK)
	{
		BeginScene(tex);
	}
	gb2D::~gb2D()
	{
		if(m_isLocked) {
			EndScene();
		}
	}
	void gb2D::BeginScene(GBCOLOR* data, int width, int height, int pitch) {
		if(m_isLocked) {
			throw( Ghulbus::gbException( Ghulbus::gbException::GB_INVALIDCONTEXT ) );
		}
		if( (!data) || (pitch < 0) ) {
			throw( Ghulbus::gbException( Ghulbus::gbException::GB_ILLEGALPARAMETER ) );
		}
		m_Data     = data;
		m_width    = width;
		m_height   = height;
		m_pitch    = pitch;
		m_isLocked = true;
	}

	void gb2D::BeginScene(gbTexture* tex) {
		if(m_isLocked) {
			throw( Ghulbus::gbException( Ghulbus::gbException::GB_INVALIDCONTEXT ) );
		}
		if(tex->IsLocked()) {
			throw( Ghulbus::gbException( Ghulbus::gbException::GB_ILLEGALPARAMETER ) );
		}
		tex->LockTexture(&m_Data, &m_pitch);
		m_Texture  = tex;
		m_width    = m_Texture->GetWidth();
		m_height   = m_Texture->GetHeight();
		m_isLocked = true;
	}

	void gb2D::EndScene() {
		if(!m_isLocked) {
			throw( Ghulbus::gbException( Ghulbus::gbException::GB_INVALIDCONTEXT ) );
		}
		if(m_Texture) {  m_Texture->UnlockTexture(); m_Texture = NULL;  }
		m_Data     = NULL;
		m_width    = -1;
		m_height   = -1;
		m_pitch    = -1;
		m_isLocked = false;
	}

	bool gb2D::IsLocked() {
		return m_isLocked;
	}
	int gb2D::GetWidth() {
		return m_width;
	}
	int gb2D::GetHeight() {
		return m_height;
	}

	Ghulbus::gbException::GBERROR gb2D::GetErrorState() {
		Ghulbus::gbException::GBERROR ret = m_errorstate;
		m_errorstate = Ghulbus::gbException::GB_OK;
		return ret;
	}

	void gb2D::SetClearColor(GBCOLOR c) {
		m_clearcolor = c;
	}
	void gb2D::TestLock() {
		if(!m_isLocked) {
			throw( Ghulbus::gbException( Ghulbus::gbException::GB_INVALIDCONTEXT ) );
		}
	}
	void gb2D::Clear() {
		TestLock();
		for(int y=0; y<m_height; ++y) {
			for(int x=0; x<m_width; ++x) {
				m_Data[y * m_pitch + x] = m_clearcolor;
			}
		}
	}
	void gb2D::PutPixel(int x, int y, GBCOLOR c)
	{
		TestLock();
		if( (x<0) || (x > (m_width-1)) || (y<0) || (y > (m_height-1)) ) {
			m_errorstate = Ghulbus::gbException::GB_ILLEGALPARAMETER;
			return;
		}
		m_Data[y * m_pitch + x] = c;
	}
	void gb2D::Line(int startX, int startY, int endX, int endY, GBCOLOR color)
	{
		TestLock();
		int bres_lookup[8][4] = 
			{{  0,-1,-1,-1 },{  0,-1, 1,-1 },{  0, 1,-1, 1 },{  0, 1, 1, 1 },
		     { -1, 0,-1,-1 },{  1, 0, 1,-1 },{ -1, 0,-1, 1 },{  1, 0, 1, 1 } };
		int bres_dx, bres_dy, bres_da, bres_db, bres_grad;
		bool Z = true;
		//Bresenham's Line:
		bres_dx = endX - startX;		//line width
		bres_dy = endY - startY;		//line height
		bres_da = abs(bres_dx); bres_db = abs(bres_dy);

		if(bres_da-bres_db < 0) {
			int tmp = bres_da;
			bres_da = bres_db;
			bres_db = tmp;
			Z = false;
		}
		int *m = bres_lookup[((bres_dx>=0)?1:0) + ((bres_dy>=0)?2:0) + (Z?4:0)];
		bres_grad = 2* bres_db - bres_da;
		PutPixel(startX, startY, color);
		for(int i = 0; i < bres_da; i++) {
			if(bres_grad >= 0) {
				startX += m[2];
				startY += m[3];
				bres_grad += 2 * bres_db - 2 * bres_da;
			} else {
				startX += m[0];
				startY += m[1];
				bres_grad += 2 * bres_db;
			}
			PutPixel(startX, startY, color);
		}
	}
	void gb2D::HLine(int startX, int endX, int y, GBCOLOR color)
	{
		TestLock();
		//Clipping:
		if( (startX > m_width-1) || (endX<0) || (y<0) || (y>m_height-1) || (startX > endX) ) {
			m_errorstate = Ghulbus::gbException::GB_ILLEGALPARAMETER;
			return;
		}
		if(startX < 0)       { startX = 0;         }
		if(endX > m_width-1) { endX   = m_width-1; }
		//Drawing:
		GBCOLOR* offset = &(m_Data[startX + (y * m_pitch)]);
		for(int i=startX; i<=endX; i++) {
			*(offset++) =  color;
		}
	}
	void gb2D::HLineGradient(int startX, int endX, int y, GBCOLOR c1, GBCOLOR c2)
	{
		TestLock();

		//Color arithmetics:
		int r1 = GBCOLOR32::GetR(c1);
		int g1 = GBCOLOR32::GetG(c1);
		int b1 = GBCOLOR32::GetB(c1);
		int a1 = GBCOLOR32::GetA(c1);
		int r2 = GBCOLOR32::GetR(c2);
		int g2 = GBCOLOR32::GetG(c2);
		int b2 = GBCOLOR32::GetB(c2);
		int a2 = GBCOLOR32::GetA(c2);
		int dX = (endX - startX);
		int gr, gg, gb, ga;			//color gradients for rgb components
		gr = ((r2 - r1)<<16)/dX;
		gg = ((g2 - g1)<<16)/dX;
		gb = ((b2 - b1)<<16)/dX;
		ga = ((a2 - a1)<<16)/dX;

		//Clipping:
		int i = 0;
		if( (startX > m_width-1) || (endX<0) || (y<0) || (y>m_height-1) || (startX > endX) ) { 
			m_errorstate = Ghulbus::gbException::GB_ILLEGALPARAMETER;
			return;
		}
		if(startX < 0) { i = -startX; startX = 0; }
		if(endX > m_width-1) { dX -= (endX-m_width-1); endX = m_width-1; }

		//Drawing:
		GBCOLOR* offset = &(m_Data[startX + (y*m_pitch)]);
		for(; i<=dX; i++) {
			*(offset++) =  GBCOLOR32::ARGB((a1 + ((ga * i)>>16) ), 
										   (r1 + ((gr * i)>>16) ),
										   (g1 + ((gg * i)>>16) ),
										   (b1 + ((gb * i)>>16) ) );
		}
	}

	void gb2D::Circle(int posX, int posY, int radius, GBCOLOR color) 
	{
		TestLock();

		int f = 1 - radius;
		int ddF_x = 0;  
		int ddF_y = -2 * radius;
		int x = 0;  
		int y = radius;
	  
		PutPixel(posX, posY + radius, color);
		PutPixel(posX, posY - radius, color);
		PutPixel(posX + radius, posY, color);
		PutPixel(posX - radius, posY, color);
	 
		while(x < y) {
			if(f >= 0) {
				y--;
				ddF_y += 2;
				f += ddF_y;
			}
			x++;
			ddF_x += 2;
			f += ddF_x + 1;
	 
			PutPixel(posX + x, posY + y, color);
			PutPixel(posX - x, posY + y, color);
			PutPixel(posX + x, posY - y, color);
			PutPixel(posX - x, posY - y, color);
			PutPixel(posX + y, posY + x, color);
			PutPixel(posX - y, posY + x, color);
			PutPixel(posX + y, posY - x, color);
			PutPixel(posX - y, posY - x, color);
		}
	}

	void gb2D::Rectangle(int left, int top, int right, int bottom, GBCOLOR color)
	{
		TestLock();
		//Clipping:
		if((left >= right) || (top >= bottom)) { 
			m_errorstate = Ghulbus::gbException::GB_ILLEGALPARAMETER;
			return; 
		}
		bool draw_l, draw_r, draw_t, draw_b;
		if(left < 0)             { left = 0;             draw_l = false; } else { draw_l = true; }
		if(top < 0)              { top = 0;              draw_t = false; } else { draw_t = true; }
		if(right > m_width-1)    { right  = m_width-1;   draw_r = false; } else { draw_r = true; }
		if(bottom > m_height-1)  { bottom = m_height-1;  draw_b = false; } else { draw_b = true; }

		//Drawing:
		GBCOLOR *offset;
		//draw top:
		if(draw_t) {
			offset = &(m_Data[left + (top*m_pitch)]);
			for(int i=left; i<=right; i++) {
				*(offset++) = color; 
			}
		}
		//draw bottom:
		if(draw_b) {
			offset = &(m_Data[left + (bottom*m_pitch)]);
			for(int i=left; i<=right; i++) {
				*(offset++) = color;
			}
		}
		//draw left:
		if(draw_l) {
			for(int i=top+1; i<bottom; i++) {
				offset = &(m_Data[left + (i*m_pitch)]);
				*offset = color;
			}
		}
		//draw right:
		if(draw_r) {
			for(int i=top+1; i<bottom; i++) {
				offset = &(m_Data[right + (i*m_pitch)]);
				*offset = color;
			}
		}
	}

	void gb2D::RectangleGradient(int left, int top, int right, int bottom, 
								 GBCOLOR ul, GBCOLOR ur, GBCOLOR ll, GBCOLOR lr)
	{
		TestLock();

		//Clipping is left to the called drawing functions HLineGradient() and PutPixel()

		//Color interpolation requires bilinear interpolation;
		// First we calculate the gradients for the left and right borders:
		int r_tl = GBCOLOR32::GetR(ul);	//top left
		int g_tl = GBCOLOR32::GetG(ul);
		int b_tl = GBCOLOR32::GetB(ul);
		int a_tl = GBCOLOR32::GetA(ul);
		int r_bl = GBCOLOR32::GetR(ll);	//bottom left
		int g_bl = GBCOLOR32::GetG(ll);
		int b_bl = GBCOLOR32::GetB(ll);
		int a_bl = GBCOLOR32::GetA(ll);
		int r_tr = GBCOLOR32::GetR(ur);	//top right
		int g_tr = GBCOLOR32::GetG(ur);
		int b_tr = GBCOLOR32::GetB(ur);
		int a_tr = GBCOLOR32::GetA(ur);
		int r_br = GBCOLOR32::GetR(lr);	//bottom right
		int g_br = GBCOLOR32::GetG(lr);
		int b_br = GBCOLOR32::GetB(lr);
		int a_br = GBCOLOR32::GetA(lr);

		int dX = (bottom - top);
		int gr_l, gg_l, gb_l, ga_l;			//color gradients left side
		gr_l = ((r_bl - r_tl)<<16)/dX;
		gg_l = ((g_bl - g_tl)<<16)/dX;
		gb_l = ((b_bl - b_tl)<<16)/dX;
		ga_l = ((a_bl - a_tl)<<16)/dX;
		int gr_r, gg_r, gb_r, ga_r;			//color gradients right side
		gr_r = ((r_br - r_tr)<<16)/dX;
		gg_r = ((g_br - g_tr)<<16)/dX;
		gb_r = ((b_br - b_tr)<<16)/dX;
		ga_r = ((a_br - a_tr)<<16)/dX;

		//Drawing:
		GBCOLOR c_l, c_r;
		//top:
		c_l = GBCOLOR32::ARGB(a_tl, r_tl, g_tl, b_tl);
		c_r = GBCOLOR32::ARGB(a_tr, r_tr, g_tr, b_tr);
		HLineGradient(left, right, top, c_l, c_r);
		//bottom:
		c_l = GBCOLOR32::ARGB(a_bl, r_bl, g_bl, b_bl);
		c_r = GBCOLOR32::ARGB(a_br, r_br, g_br, b_br);
		HLineGradient(left, right, bottom, c_l, c_r);

		for(int i=0; i<=dX; i++) {
			c_l = GBCOLOR32::ARGB((a_tl + ((ga_l * i)>>16) ),
								  (r_tl + ((gr_l * i)>>16) ),
								  (g_tl + ((gg_l * i)>>16) ),
								  (b_tl + ((gb_l * i)>>16) ) );
			c_r = GBCOLOR32::ARGB((a_tr + ((ga_r * i)>>16) ),
								  (r_tr + ((gr_r * i)>>16) ),
								  (g_tr + ((gg_r * i)>>16) ),
								  (b_tr + ((gb_r * i)>>16) ) );
			PutPixel(left, top+i, c_l);
			PutPixel(right, top+i, c_r);
		}
	}

	void gb2D::Box(int left, int top, int right, int bottom, GBCOLOR color)
	{
		TestLock();

		//Clipping:
		if(left < 0) { left = 0; }
		if(top < 0)	 { top = 0; }
		if(right > m_width-1)   { right = m_width-1; }
		if(bottom > m_height-1) { bottom = m_height-1; }
		//Drawing:
		for(int i=top; i<=bottom; i++) {
			GBCOLOR* offset = &(m_Data[left + (i*m_pitch)]);
			for(int j=left; j<=right; j++) {
				*(offset++) =  color;
			}
		}
	}

	void gb2D::BoxGradient(int left, int top, int right, int bottom, 
						   GBCOLOR ul, GBCOLOR ur, GBCOLOR ll, GBCOLOR lr)
	{
		TestLock();

		//Clipping is left to the called drawing function HLineGradient()

		//Color interpolation requires bilinear interpolation;
		// First we calculate the gradients for the left and right borders:
		int r_tl = GBCOLOR32::GetR(ul);	//top left
		int g_tl = GBCOLOR32::GetG(ul);
		int b_tl = GBCOLOR32::GetB(ul);
		int a_tl = GBCOLOR32::GetA(ul);
		int r_bl = GBCOLOR32::GetR(ll);	//bottom left
		int g_bl = GBCOLOR32::GetG(ll);
		int b_bl = GBCOLOR32::GetB(ll);
		int a_bl = GBCOLOR32::GetA(ll);
		int r_tr = GBCOLOR32::GetR(ur);	//top right
		int g_tr = GBCOLOR32::GetG(ur);
		int b_tr = GBCOLOR32::GetB(ur);
		int a_tr = GBCOLOR32::GetA(ur);
		int r_br = GBCOLOR32::GetR(lr);	//bottom right
		int g_br = GBCOLOR32::GetG(lr);
		int b_br = GBCOLOR32::GetB(lr);
		int a_br = GBCOLOR32::GetA(lr);

		int dX = (bottom - top);
		int gr_l, gg_l, gb_l, ga_l;			//color gradients left side
		gr_l = ((r_bl - r_tl)<<16)/dX;
		gg_l = ((g_bl - g_tl)<<16)/dX;
		gb_l = ((b_bl - b_tl)<<16)/dX;
		ga_l = ((a_bl - a_tl)<<16)/dX;
		int gr_r, gg_r, gb_r, ga_r;			//color gradients right side
		gr_r = ((r_br - r_tr)<<16)/dX;
		gg_r = ((g_br - g_tr)<<16)/dX;
		gb_r = ((b_br - b_tr)<<16)/dX;
		ga_r = ((a_br - a_tr)<<16)/dX;

		//the box is drawn from top to bottom:
		GBCOLOR c_l, c_r;
		for(int i=0; i<=dX; i++) {
			c_l = GBCOLOR32::ARGB((a_tl + ((ga_l * i)>>16) ),
								  (r_tl + ((gr_l * i)>>16) ),
								  (g_tl + ((gg_l * i)>>16) ),
								  (b_tl + ((gb_l * i)>>16) ) );
			c_r = GBCOLOR32::ARGB((a_tr + ((ga_r * i)>>16) ),
								  (r_tr + ((gr_r * i)>>16) ),
								  (g_tr + ((gg_r * i)>>16) ),
								  (b_tr + ((gb_r * i)>>16) ) );
			//second interpolation step as well as clipping is left to HLineGradient():
			HLineGradient(left, right, top+i, c_l, c_r);
		}
	}

}

