#include "../include/gbGraphics.hpp"

namespace GhulbusGraphics {
#ifdef _GHULBUSGRAPHICS__USE_SDL_

	SDL_Surface* SetGraphicsMode(int width, int height, bool fullscreen) {
		//request 32bit RGBA color mode:
		SDL_GL_SetAttribute(SDL_GL_BUFFER_SIZE, 32);
		SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE,   8);
		SDL_GL_SetAttribute(SDL_GL_RED_SIZE,     8);
		SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE,   8);
		SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,    8);

		//request 24bit Z-buffer and 8bit Stencil buffer:
		SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,  24);
		SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 8);
		//request double buffering:
		SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
		
		return SDL_SetVideoMode(width, height, 32,
								SDL_OPENGL | SDL_OPENGLBLIT | (fullscreen?SDL_FULLSCREEN:0x00000000));
	}
	bool CheckGraphicsMode(SDL_Surface* screen, int width, int height, bool fullscreen) {
		if(!screen) {
			return false;
		}
		unsigned int tmp_flags = (SDL_OPENGL | SDL_OPENGLBLIT | (fullscreen?SDL_FULLSCREEN:0x00000000));
		if( (screen->w != width) || (screen->h != height) || ((screen->flags & tmp_flags) != tmp_flags) ) {
			return false;
		}
		int tmp_attr;
		SDL_GL_GetAttribute(SDL_GL_BUFFER_SIZE,  &tmp_attr);
		if(tmp_attr < 32)  { return false; }
		SDL_GL_GetAttribute(SDL_GL_ALPHA_SIZE,   &tmp_attr);
		if(tmp_attr != 8)  { return false; }
		SDL_GL_GetAttribute(SDL_GL_RED_SIZE,     &tmp_attr);
		if(tmp_attr != 8)  { return false; }
		SDL_GL_GetAttribute(SDL_GL_GREEN_SIZE,   &tmp_attr);
		if(tmp_attr != 8)  { return false; }
		SDL_GL_GetAttribute(SDL_GL_BLUE_SIZE,    &tmp_attr);
		if(tmp_attr != 8)  { return false; }
		SDL_GL_GetAttribute(SDL_GL_DEPTH_SIZE,   &tmp_attr);
		if(tmp_attr != 24) { return false; }
		SDL_GL_GetAttribute(SDL_GL_STENCIL_SIZE, &tmp_attr);
		if(tmp_attr != 8)  { return false; }
		return true;
	}

	void gbGraphicsManager_SDL::InitOpenGL() {
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glShadeModel(GL_SMOOTH);
		glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
		glFrontFace(GL_CCW);
		glCullFace(GL_BACK);
		glEnable(GL_CULL_FACE);
		SetClearColor(m_clearcolor);
		glClearDepth(1.0);
		glEnable(GL_DEPTH_TEST);
		glDepthFunc(GL_LEQUAL);
		glEnable(GL_TEXTURE_2D);
	}

	void gbGraphicsManager_SDL::LoadOpenGLExtensions() {
		GLenum err = glewInit();
		if (GLEW_OK != err) {
			throw( Ghulbus::gbException( Ghulbus::gbException::GB_FAILED,
										 "Unable to initialize GLEW" ) );
		}
		std::cout << "GLEW up and running - " << glewGetString(GLEW_VERSION) << std::endl;
		if(GLEW_ARB_texture_non_power_of_two) {
			std::cout << " * GLEW_ARB_texture_non_power_of_two is supported" << std::endl;
			m_npot_textures_supported = true;
		} else {
			std::cout << " WARNING - Missing OpenGL extension for arbitrary sized textures" << std::endl;
			m_npot_textures_supported = false;

		}
	}

	void gbGraphicsManager_SDL::BuildOverlay() {
		GLuint id;
		glGenTextures(1, &id);
		glBindTexture(GL_TEXTURE_2D, id);
		m_Overlay = new gbTexture_SDL(id, m_scrWidth, m_scrHeight, m_npot_textures_supported);
											//will be deleted by destructor of base class gbGraphicsManager
	}

	void gbGraphicsManager_SDL::RenderOverlay() {
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		glEnable(GL_TEXTURE_2D);
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glBindTexture(GL_TEXTURE_2D, dynamic_cast<gbTexture_SDL*>(m_Overlay)->GetTextureId());
		glDisable(GL_DEPTH_TEST);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);

		float texCoordU = 1.0f;
		float texCoordV = 1.0f;
		if(!m_npot_textures_supported) {
			texCoordU = ( static_cast<float>(m_Overlay->GetWidth()) / 
						  static_cast<float>(dynamic_cast<gbTexture_SDL*>(m_Overlay)->GetTextureWidth()) );
			texCoordV = ( static_cast<float>(m_Overlay->GetHeight()) / 
						  static_cast<float>(dynamic_cast<gbTexture_SDL*>(m_Overlay)->GetTextureHeight()) );
		}
		glBegin(GL_TRIANGLE_STRIP);
			glColor4f(1.0, 1.0, 1.0, 1.0);
			glTexCoord2f(0.0,       texCoordV);  glVertex3f(-1.0, -1.0,  0.0);		//LL
			glTexCoord2f(texCoordU, texCoordV);  glVertex3f( 1.0, -1.0,  0.0);		//LR
			glTexCoord2f(0.0,       0.0);        glVertex3f(-1.0,  1.0,  0.0);		//UL
			glTexCoord2f(texCoordU, 0.0);        glVertex3f( 1.0,  1.0,  0.0);		//UR
		glEnd();

		glDisable(GL_BLEND);
		glEnable(GL_DEPTH_TEST);

		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
	}

	gbGraphicsManager_SDL::gbGraphicsManager_SDL(int scrWidth, int scrHeight, bool fullscreen)
		:m_Screen(NULL), m_glError(GL_NO_ERROR), m_npot_textures_supported(false)
	{
		if( !SDL_WasInit(SDL_INIT_VIDEO) ) {
			if( SDL_Init( SDL_INIT_VIDEO ) < 0 ) {
				throw( Ghulbus::gbException(Ghulbus::gbException::GB_FAILED,
											"Unable to initialize SDL") );
			}
		}
		///@todo unfortunately, no multi-window support by sdl until now :(
		m_Screen = SetGraphicsMode(scrWidth, scrHeight, fullscreen);
		if( !CheckGraphicsMode(m_Screen, scrWidth, scrHeight, fullscreen) ) {
			throw( Ghulbus::gbException(Ghulbus::gbException::GB_FAILED,
				                        "Unable to set Video Mode") );
		}
		m_scrWidth     = scrWidth;
		m_scrHeight    = scrHeight;
		m_isFullscreen = fullscreen;
		LoadOpenGLExtensions();
		InitOpenGL();
		BuildOverlay();
	}

	gbGraphicsManager_SDL::~gbGraphicsManager_SDL() {
		for(std::vector<gbGraphicsObject*>::iterator i = m_ObjectList.begin(); i != m_ObjectList.end(); ++i) {
			(*i)->Release();
		}
		//release overlay texture:
		GLuint id = dynamic_cast<gbTexture_SDL*>(m_Overlay)->GetTextureId();
		glDeleteTextures(1, &id);
		SDL_Quit();
	}

	void gbGraphicsManager_SDL::BeginScene()
	{
		SDL_GL_SwapBuffers();
		if(m_isRendering) {
			throw( Ghulbus::gbException( Ghulbus::gbException::GB_INVALIDCONTEXT ) );
		}
		glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
		m_isRendering = true;
	}
	void gbGraphicsManager_SDL::EndScene()
	{
		if(!m_isRendering) {
			throw( Ghulbus::gbException( Ghulbus::gbException::GB_INVALIDCONTEXT ) );
		}
		for(std::vector<gbGraphicsObject*>::iterator i = m_ObjectList.begin(); i != m_ObjectList.end(); ++i) {
			(*i)->Render();
		}
		RenderOverlay();
		m_isRendering = false;
	}
	void gbGraphicsManager_SDL::Clear()
	{
		if(!m_isRendering) {
			throw( Ghulbus::gbException( Ghulbus::gbException::GB_INVALIDCONTEXT ) );
		}
		glClear(GL_COLOR_BUFFER_BIT);
	}
	void gbGraphicsManager_SDL::SetClearColor(GBCOLOR c)
	{
		m_clearcolor = c;
		glClearColor( GBCOLOR32::GetR(m_clearcolor) / 255.0f,
					  GBCOLOR32::GetG(m_clearcolor) / 255.0f,
					  GBCOLOR32::GetB(m_clearcolor) / 255.0f,
					  GBCOLOR32::GetA(m_clearcolor) / 255.0f );
	}
	void gbGraphicsManager_SDL::Render(unsigned int dT)
	{
	}

	gbSprite* gbGraphicsManager_SDL::CreateSprite(int width, int height) {
		gbSprite_SDL* ret;
		gbTexture_SDL* tTexture;
		GLuint id;

		glGenTextures(1, &id);
		glBindTexture(GL_TEXTURE_2D, id);	//binding ensures id is a 2d texture

		tTexture = new gbTexture_SDL(id, width, height, m_npot_textures_supported);	
																//will be deleted by gbSprite destructor
		ret = new gbSprite_SDL(tTexture, m_scrWidth, m_scrHeight);
		m_ObjectList.push_back(ret);

		return ret;
	}

	void gbGraphicsManager_SDL::ProcessMessages()
	{
		SDL_Event msg;
		while(SDL_PollEvent(&msg)) {
			switch(msg.type) {
				case SDL_KEYDOWN:
					switch(msg.key.keysym.sym) {
						case SDLK_ESCAPE:
							memset(&msg, 0, sizeof(SDL_Event));
							msg.type = SDL_QUIT;
							SDL_PushEvent(&msg);
						default:
							break;
					}
					break;
				case SDL_QUIT:
					SetDone();
					break;
				default:
					break;
			}
		}
	}
#endif
}

